/* ----------------------------------------------------------------------
cmain.c : main program for testing esf arrays
John O'Donnell, March 2013
---------------------------------------------------------------------- */

#include <stdio.h>
#include "esftypes.h"

/* ----------------------------------------------------------------------
			       Example
---------------------------------------------------------------------- */

/* This example procedure illustrates how to use esf arrays in your
   program.

   See the comments near the beginning of esf.c for a description of
   the operations that are supported.

 * You need to compile your program with esftypes.h and esf.c.
      cc -c esf.c
      cc -o cmain.exe main.c esf.o
      ./cmain

 * There are two #define statements in esf.c that you might want to
   adjust.  ESF_NCELLS is the size of the memory, and ESF_TRACE says
   whether you want the esf operations to print some internal tracing
   information.  It's recommended to set ESF_NCELLS to a reasonable
   power of 2, and set ESF_TRACE to 0.

 * Define your esf arrays to have either type int or esf_array.

 * The indices and values of esf arrays have type int.

 * Before doing any esf operations, call esfc_init().

 * The empty array always exists, and its name is ESF_EMPTY.  Don't
   use a numeric literal for this; use ESF_EMPTY.

 * Suppose you want to create a newarray which is the same as an
   existing oldarray, except that newarray[i]=x. Do this with update:
   newarray = esfc_update (oldarray, i, x);

 * To get the value of a[i], use esfc_lookup (a,i).

 * The functions that start with esfp are the primitive operations,
   while functions that start with esfc do some checking, and if the
   esf operation failed the esfc_function will terminate the program.

 * Many of the esf operations can fail.  For example, you could do a
   lookup to an element that doesn't exist.  The esfp_functions
   generally return a pointer to a struct that contains a status
   indicator and the result.  However, it's often more convenient just
   to terminate the program if an operation fails, so you don't have
   to check the return status for every operation.  That's the purpose
   of the esfc_operations: they call the primitive esfp_operation,
   check the return status, abort the program in the event of failure,
   and return the result as a simple int.

*/

void example ()
{
  esf_array a1, a2;
  printf ("Running example\n");
  esfc_init ();   /* initialize the esf machine */

  a1 = esfc_update (ESF_EMPTY, 1, 10);
  printf ("a1[1] =%d\n", esfc_lookup (a1,1));

  a2 = esfc_update (a1, 2, 20);
  printf ("a2[1] =%d\n", esfc_lookup (a2,1));
  printf ("a2[2] =%d\n", esfc_lookup (a2,2));

/* The following test will terminate the program, because a2[3] is
   undefined
  printf ("a2[3] =%d\n", esfc_lookup (a2,3));
*/

}

/* ----------------------------------------------------------------------
			      Basic test
---------------------------------------------------------------------- */

void basic_test ()
{
  esf_array a1, a2, a3;
  int v1, v2, v3;

  printf ("Running basic_test\n");

  esfc_init ();   /* initialize the esf machine */
  esfp_dump (0);  /* print the initial state */

  a1 = esfc_update (ESF_EMPTY, 1, 10);
    /* a1 = 1->10 */
  printf ("name a1 = %d\n", a1);
  esfp_dump(0);

  a2 = esfc_update (a1, 2, 20);
    /* a2 = 1->10, 2->20  */
  printf ("name a2 = %d\n", a2);
  esfp_dump(0);

  a3 = esfc_update (a2, 3, 30);
    /* a2 = 1->10, 2->20, 3->30  */
  printf ("name a3 = %d\n", a3);
  esfp_dump(0);

  v1 = esfc_lookup (a3, 1);  /* 10 */
  v2 = esfc_lookup (a3, 2);  /* 20 */
  v3 = esfc_lookup (a3, 3);  /* 30 */
  printf ("v1=%d, v2=%d, v3 = %d\n", v1,v2,v3);

  printf ("Final machine state:\n");
  esfp_dump (0);

/* The following test will terminate the program...
  printf ("Try looking up a nonexistent array element\n");
  v1 = esfc_lookup (a2,3);
  printf ("v1=%d\n", v1);
*/
}

/* ----------------------------------------------------------------------
		  Example used in the PARLE'93 paper
---------------------------------------------------------------------- */

void parle_test ()
{
  esf_array a1, a2, a3, a4, a5, a6;
  int x1, x2, x3;

  printf ("Running parle_test\n");

  esfc_init ();   /* initialize the esf machine */
  esfp_dump (0);  /* print the initial state */

  a1 = esfc_update (ESF_EMPTY, 100, 101);
  a2 = esfc_update (a1, 200, 201);
  a3 = esfc_update (a1, 300, 301);
  a4 = esfc_update (a2, 400, 401);
  a5 = esfc_update (a2, 500, 501);
  a6 = esfc_update (a3, 600, 601);

/*
Expected results; the notation i-v means a[i]=v.
  a1 = 100-101
  a2 = 200-201 100-101 
  a3 = 300-301 100-101
  a4 = 400-401 200-201 100-101 
  a5 = 500-501 200-201 100-101 
  a6 = 600-601 300-301 100-101
*/

  printf ("a1[100] want 101, result=%d\n", esfc_lookup (a1, 100));

  printf ("a2[100] want 101, result=%d\n", esfc_lookup (a2, 100));
  printf ("a2[200] want 201, result=%d\n", esfc_lookup (a2, 200));

  printf ("a3[100] want 101, result=%d\n", esfc_lookup (a3, 100));
  printf ("a3[300] want 301, result=%d\n", esfc_lookup (a3, 300));

  printf ("a4[100] want 101, result=%d\n", esfc_lookup (a4, 100));
  printf ("a4[200] want 201, result=%d\n", esfc_lookup (a4, 200));
  printf ("a4[400] want 401, result=%d\n", esfc_lookup (a4, 400));

  printf ("a5[100] want 101, result=%d\n", esfc_lookup (a5, 100));
  printf ("a5[200] want 201, result=%d\n", esfc_lookup (a5, 200));
  printf ("a5[500] want 501, result=%d\n", esfc_lookup (a5, 500));

  printf ("a6[100] want 101, result=%d\n", esfc_lookup (a6, 100));
  printf ("a6[300] want 301, result=%d\n", esfc_lookup (a6, 300));
  printf ("a6[600] want 601, result=%d\n", esfc_lookup (a6, 600));


  printf ("Final machine state:\n");
  esfp_dump (0);
}

/* ----------------------------------------------------------------------
		Extended PARLE example with shadowing
---------------------------------------------------------------------- */

/* Similar to PARLE test, except that some indices are redefined.  The
   aim of this test is to check that the right value is found for a[i]
   even if a was defined with updates providing several distinct
   elements with index i */

void parle_test_shadowing ()
{
  esf_array a1, a2, a3, a4, a5, a6;
  int x1, x2, x3;

  printf ("Running parle_test_shadowing\n");

  esfc_init ();  /* initialize the esf machine */
  esfc_init();

  esfp_dump (0);  /* print the initial state */

  a1 = esfc_update (ESF_EMPTY, 100, 101);
  a2 = esfc_update (a1, 200, 201);
  a3 = esfc_update (a1, 100, 111);  /* overrides a1[100] */
  a4 = esfc_update (a2, 400, 401);
  a5 = esfc_update (a2, 500, 501);
  a6 = esfc_update (a3, 200, 222);  /* shadows a3[200] */

/*
Expected results; the notation i-v means a[i]=v.
  a1 = 100-101
  a2 = 200-201 100-101 
  a3 = 100-111
  a4 = 400-401 200-201 100-101 
  a5 = 500-501 200-201 100-101 
  a6 = 200-222 100-111
*/

  printf ("a1[100] want 101, result=%d\n", esfc_lookup (a1, 100));

  printf ("a2[100] want 101, result=%d\n", esfc_lookup (a2, 100));
  printf ("a2[200] want 201, result=%d\n", esfc_lookup (a2, 200));

  printf ("a3[100] want 111, result=%d\n", esfc_lookup (a3, 100));

  printf ("a4[100] want 101, result=%d\n", esfc_lookup (a4, 100));
  printf ("a4[200] want 201, result=%d\n", esfc_lookup (a4, 200));
  printf ("a4[400] want 401, result=%d\n", esfc_lookup (a4, 400));

  printf ("a5[100] want 101, result=%d\n", esfc_lookup (a5, 100));
  printf ("a5[200] want 201, result=%d\n", esfc_lookup (a5, 200));
  printf ("a5[500] want 501, result=%d\n", esfc_lookup (a5, 500));

  printf ("a6[100] want 111, result=%d\n", esfc_lookup (a6, 100));
  printf ("a6[200] want 222, result=%d\n", esfc_lookup (a6, 200));

  printf ("Final machine state:\n");
  esfp_dump (0);
}


/* ----------------------------------------------------------------------
	     Traversing sparse array from max index down
---------------------------------------------------------------------- */

void sparse_test_max_down ()
{
  esf_array a1, a2, a3, a4, a5, a6;
  int i1, i2, i3, i4, i5, i6;

  printf ("Running sparse_test_max_down\n");

  esfc_init ();  /* initialize the esf machine */

  a1 = esfc_update (ESF_EMPTY, 100, 1);
  a2 = esfc_update (a1, 101, 2);
  a3 = esfc_update (a2, 129, 3);
  a4 = esfc_update (a3, 107, 5);
  a5 = esfc_update (a3, 114, 6);
  a6 = esfc_update (a5, 107, 55);
  esfp_dump (0);

  i1 = esfc_maxindex (a6);
  i2 = esfc_previndex (a6,i1);
  i3 = esfc_previndex (a6,i2);
  i4 = esfc_previndex (a6,i3);
  i5 = esfc_previndex (a6,i4);
  i6 = esfc_previndex (a6,i5);

  printf ("Indices %d %d %d %d %d %d\n", i1, i2, i3, i4, i5, i6);

}


/* ----------------------------------------------------------------------
	      Traversing sparse array from min index up
---------------------------------------------------------------------- */

void sparse_test_min_up ()
{
  esf_array a1, a2, a3, a4, a5, a6;
  int i1, i2, i3, i4, i5, i6;

  printf ("Running sparse_test_min_up\n");

  esfc_init ();  /* initialize the esf machine */

  a1 = esfc_update (ESF_EMPTY, 100, 1);
  a2 = esfc_update (a1, 101, 2);
  a3 = esfc_update (a2, 129, 3);
  a4 = esfc_update (a3, 107, 5);
  a5 = esfc_update (a3, 114, 6);
  a6 = esfc_update (a5, 107, 55);
  esfp_dump (0);

  i1 = esfc_minindex (a6);
  i2 = esfc_nextindex (a6,i1);
  i3 = esfc_nextindex (a6,i2);
  i4 = esfc_nextindex (a6,i3);
  i5 = esfc_nextindex (a6,i4);
  i6 = esfc_nextindex (a6,i5);

  printf ("Indices %d %d %d %d %d %d\n", i1, i2, i3, i4, i5, i6);

}

/* ----------------------------------------------------------------------
			     Main program
---------------------------------------------------------------------- */

main ()
{ printf ("Main program for esf, C version\n");

  example();
  printf("\n\n");

  basic_test();
  printf("\n\n");

  parle_test();
  printf("\n\n");

  parle_test_shadowing();
  printf("\n\n");

  sparse_test_max_down();
  printf("\n\n");

  sparse_test_min_up();

  printf ("Esf main program ending\n");

}
